#!/usr/bin/env python3
# -*- coding: utf-8 -*-
R"""
Some refinery units produce more than one output when applied to an input. For example,
`refinery.chop` will chop the input data into evenly sized blocks and emit each of them
as a single output. By default, if no framing syntax is used, multiple outputs are
separated by line breaks, which is often desirable when text data is extracted. However,
for processing binary data, this is equally often more than useless. To process the list
of results generated by any refinery unit, end the command for this unit with the
special argument `[`. This argument has to be the last argument to be recognized as a
framing initialization. If this syntax is used, the list of results is emitted in an
internal format which allows arbitrarily nested lists of binary chunks to be processed.

### Simple Frame Example

    $ emit OOOOOOOO | chop 2 [| ccp F | cca . ]
    FOO.FOO.FOO.FOO.

Here, the string `OOOOOOOO` is first chopped into blocks of 2, yielding the **frame**
`[OO, OO, OO, OO]` which is then forwarded to the next command. If a `refinery.units.Unit`
receives input in framed format, each chunk of the frame is processed individually and
emitted as one output chunk. In this case, `refinery.ccp` simply prepends `F` to every
input, producing the frame `[FOO, FOO, FOO, FOO]`. Finally, `refinery.cca` appends a period
to each chunk. When a unit is given the closing bracket as the last argument, this
concludes processing of one frame which results in concatenation of all binary chunks in
the frame.

### Frame Layers

Frames can be nested arbitrarily, and `refinery.sep` can be used to insert a separator
(the default is line break) between all chunks in the frame:

    $ emit OOOOOOOO | chop 4 [| chop 2 [| ccp F | cca . ]| sep ]
    FOO.FOO.
    FOO.FOO.

Here, we first produce the two-layered **frame tree** `[[OO,OO], [OO,OO]]` by using two
`refinery.chop` invocations. We refer to this data as a tree because, well, it is one:

    LAYER 1:      [[..],[..]]
                    /     \
    LAYER 2:    [OO,OO] [OO,OO]

The bottom layer is processed as before, yielding `[FOO.FOO., FOO.FOO.]`. Next, the unit
`refinery.sep` inserts a line break character between the two chunks in this frame.

### Adding Line Breaks Easily

Since separating data with line breaks is a common requirement, it is also possible to use
one more closing bracket than necessary at the end of a frame to separate all chunks by line
breaks:

    $ emit OOOOOOOO | chop 4 [| chop 2 [| ccp F | cca . ]]]
    FOO.FOO.
    FOO.FOO.

### Scoping

It is possible to alter the **visibility** of `refinery.lib.frame.Chunk`, primarily by
using `refinery.scope`. The unit accepts a slice argument which defines the indices of
the current frame that remain visible. All subsequent units will only process visible
chunks and simply forward the ones that are not visible. `refinery.lib.frame.Chunk`s
remain invisible when a new frame layer opens:

    $ emit BINARY REFINERY [| scope 0 | clower | sep - ]
    binary-REFINERY

Here, the scope was limited to the first chunk `BINARY` which was transformed to lower
case, but the second chunk `REFINERY` was left untouched. A somewhat more complex example:

    $ emit aaaaaaaa namtaB [| scope 0 | rex . [| ccp N ]| scope 1 | rev | sep - ]
    NaNaNaNaNaNaNaNa-Batman

Note that `refinery.sep` makes all chunks in the frame visible by default, because it is
intended to sit at the end of a frame. Otherwise, `NaNaNaNaNaNaNaNa` and `Batman` in the
above example would not be separated by a dash.
"""
import io
import os

from typing import Iterable, BinaryIO, Callable, Optional, Tuple, Dict, ByteString, Any

try:
    import msgpack
except ModuleNotFoundError:
    msgpack = None

__all__ = [
    'Chunk',
    'Framed',
    'FrameUnpacker'
]

MAGIC = bytes.fromhex(
    os.environ.get('REFINERY_FRAME_MAGIC', 'C0CAC01AC0DE'))


class Chunk(bytearray):
    """
    Represents the individual chunks in a frame. The `refinery.units.Unit.filter` method
    receives an iterable of `refinery.lib.frame.Chunk`s.
    """
    def __init__(
        self,
        data: ByteString,
        path: Tuple[int],
        view: Optional[Tuple[bool]] = None,
        meta: Optional[Dict[str, Any]] = None
    ):
        view = view or [False] * len(path)
        if len(view) != len(path):
            raise ValueError('skipping must have the same length as path')

        self._view = view
        self._path = path
        self._meta = meta or dict()

        bytearray.__init__(self, data)

    def nest(self, *ids):
        """
        Nest this chunk deeper by providing a sequence of indices inside each new layer of the
        frame. The `refinery.lib.frame.Chunk.path` tuple is extended by these values. The
        visibility of the `refinery.lib.frame.Chunk` at each new layer is inherited from its
        current visibility.
        """
        self._view += (self.visible,) * len(ids)
        self._path += ids
        return self

    @property
    def view(self) -> Tuple[bool]:
        """
        This tuple of boolean values indicates the visibility of this chunk at each layer of
        the frame tree. The `refinery.scope` unit can be used to change visibility of chunks
        within a frame.
        """
        return self._view

    @property
    def path(self) -> Tuple[int]:
        """
        The vertices in each frame tree layer are sequentially numbered by their order of
        appearance in the stream. The `refinery.lib.frame.Chunk.path` contains the numbers of
        the vertices (in each layer) which define the path from the root of the frame tree
        to the leaf vertex representing this `refinery.lib.frame.Chunk`
        """
        return self._path

    @property
    def meta(self) -> Dict[str, Any]:
        """
        Every chunk can contain a dictionary of arbitrary metadata.
        """
        return self._meta

    @property
    def visible(self):
        """
        This property defines whether the chunk is currently visible. It defaults to true if the
        chunk is not part of a frame and is otherwise the same as the last element of the tuple
        `refinery.lib.frame.Chunk.view`. Setting this property will correspondingly alter the last
        entry of `refinery.lib.frame.Chunk.view`.
        Setting this property on an unframed `refinery.lib.frame.Chunk` raises an `AttributeError`.
        """
        return not self._view or self._view[~0]

    @property
    def scopable(self):
        """
        This property defines whether the chunk can be made visible in the current frame.
        """
        return len(self._view) <= 1 or self._view[~1]

    @visible.setter
    def visible(self, value: bool):
        if not self._view:
            raise AttributeError('cannot set visibility of chunk outside frame')
        if value != self.visible:
            self._view = self._view[:~0] + (value,)

    @classmethod
    def unpack(cls, stream):
        """
        Classmethod to read a serialized chunk from an unpacker stream.
        """
        path, view, meta, data = next(stream)
        return cls(data, path, view=view, meta=meta)

    def pack(self):
        """
        Return the serialized representation of this chunk.
        """
        return msgpack.packb((self._path, self._view, self._meta, self))

    def __repr__(self) -> str:
        layer = '/'.join('#' if not s else str(p)
            for p, s in zip(self._path, self._view))
        layer = layer and '/' + layer
        metas = ' '.join(F'{key}={value!s}' for key, value in self._meta.items())
        metas = metas and ' ' + metas
        return F'<chunk{layer} size={len(self)}{metas}>'

    def __hash__(self):
        return hash((
            len(self),
            bytes(self[:+64]),
            bytes(self[-64:])
        ))

    def __getitem__(self, bounds):
        if isinstance(bounds, str):
            return self._meta.get(bounds, None)
        return bytearray.__getitem__(self, bounds)

    def __setitem__(self, bounds, value):
        if isinstance(bounds, str):
            self._meta[bounds] = value
        else:
            bytearray.__setitem__(self, bounds, value)


class FrameUnpacker:
    """
    Provides a unified interface to read both framed and raw input data from a stream. After
    loading a framed input stream, the object provides an iterator over the first **frame** in
    the bottom **layer** of the frame tree. Consider this doubly layered frame tree:

        [[FOO, BAR], [BOO, BAZ]]

    The `refinery.lib.frame.FrameUnpacker` object will first be an iterator over the first frame
    `[FOO, BAR]`. After consuming this iterator, the `refinery.lib.frame.FrameUnpacker.nextframe`
    method can be called to load the next frame, at which point the object will become an
    iterator over `[BOO, BAZ]`.
    """
    def __init__(self, stream: Optional[BinaryIO]):
        import msgpack
        self.finished = False
        self.trunk = ()
        self._next = Chunk(bytearray(), ())
        buffer = stream and stream.read(len(MAGIC)) or B''
        if buffer == MAGIC:
            self.framed = True
            self.stream = stream
            self.unpacker = msgpack.Unpacker(
                max_buffer_size=0xFFFFFFFF,
                use_list=False
            )
            self._advance()
            self.gauge = len(self._next.path)
        else:
            self.framed = False
            self.gauge = 0
            while buffer:
                self._next.extend(buffer)
                buffer = stream.read()

    def _advance(self) -> None:
        while not self.finished:
            try:
                self._next = Chunk.unpack(self.unpacker)
                break
            except StopIteration:
                pass
            try:
                recv = self.stream.read1() or self.stream.read()
            except TypeError:
                recv = self.stream.read()
            if recv:
                self.unpacker.feed(recv)
                continue
            self.finished = True

    def nextframe(self) -> bool:
        """
        Once the iterator is consumed, calling this function will return `True` if
        and only if another frame with input data has been loaded, in which case
        the object will provide an iterator over the freshly loaded frame. If this
        function returns `False`, all input data has been consumed.
        """
        if self.finished:
            return False
        self.trunk = self._next.path
        return True

    @property
    def eol(self) -> bool:
        return self.trunk != self.peek

    @property
    def peek(self) -> Tuple[int]:
        """
        Contains the identifier of the next frame.
        """
        return self._next.path

    def __iter__(self) -> Iterable[Chunk]:
        if self.finished:
            return
        if not self.framed:
            yield self._next
            self.finished = True
            return
        while not self.finished and self.trunk == self._next.path:
            yield self._next
            self._advance()


class Framed:
    """
    A proxy interface to ingest and output framed data. It is given an `action` to be
    performed for each elementary chunk of data, a `stream` of input data, and an integer
    argument `nested` which specifies the relative amount of nesting to be performed
    by the interface. This parameter should either be `1` if the interface should output
    the results at an additional layer, `0` if the nesting depth of the data should
    remain unchanged, and a negative amount if frame layers are to be collapsed. After
    initialization, the `refinery.lib.frame.Framed` object is an iterator that yields
    bytestrings which can be forwarded as the output of the operation with all framing
    already taken care of.
    """
    def __init__(
        self,
        action: Callable[[bytearray], Iterable[bytes]],
        stream: BinaryIO,
        nested: int = 0,
        filter: Optional[Callable[[Iterable[Chunk]], Iterable[Chunk]]] = None,
    ):
        self.unpack = FrameUnpacker(stream)
        self.action = action
        self.nested = nested
        self.filter = filter

    def _apply_filter(self) -> Iterable[Chunk]:

        it = iter(self.unpack)

        def rewind():
            yield top
            yield from it

        try:
            top = next(it)
        except StopIteration:
            pass
        else:
            rw = rewind()
            yield from self.filter(rw) if top.scopable else rw

        if not self.unpack.eol:  # filter did not consume the iterable, abort
            self.unpack.finished = True

    @property
    def unframed(self) -> bool:
        """
        This property is true if the output data is not framed.
        """
        return self.nested < 1 and not self.unpack.framed

    @property
    def framebreak(self) -> bool:
        """
        This property will be true if the data generated by this framing interface
        is unframed, and the requested nesting was smaller than required to achieve
        this. In practice, it means that the user has provided more closing brakcets
        than were required to close all open frames.
        """
        return self.nested + self.unpack.gauge < 0

    def _generate_chunks(self, parent: Chunk):
        for item in self.action(parent):
            meta = parent.meta
            data = item
            if isinstance(data, dict):
                data = item.pop('data', None)
                meta = item
            yield Chunk(data, parent.path, parent.view, meta)

    def _generate_bytes(self, data: ByteString):
        for item in self.action(data):
            if isinstance(item, dict):
                item = item.pop('data', None)
            yield item

    def __iter__(self):
        if self.unpack.finished:
            return

        if self.nested > 0:
            yield MAGIC
            if self.nested > 1:
                rest = (0,) * (self.nested - 2)
                while self.unpack.nextframe():
                    for k, chunk in enumerate(self._apply_filter()):
                        if not chunk.visible:
                            yield chunk.nest(k, 0, *rest).pack()
                            continue
                        for j, result in enumerate(self._generate_chunks(chunk)):
                            yield result.nest(k, j, *rest).pack()
            else:
                while self.unpack.nextframe():
                    for k, chunk in enumerate(self._apply_filter()):
                        if not chunk.visible:
                            yield chunk.nest(k).pack()
                            continue
                        for result in self._generate_chunks(chunk):
                            yield result.nest(k).pack()
            return

        elif not self.unpack.framed:
            for chunk in self._apply_filter():
                yield from self._generate_bytes(chunk)
            return

        if self.nested == 0:
            yield MAGIC
            while self.unpack.nextframe():
                for chunk in self._apply_filter():
                    if not chunk.visible:
                        yield chunk.pack()
                        continue
                    for result in self._generate_chunks(chunk):
                        yield result.pack()
            return

        gauge = max(self.unpack.gauge + self.nested, 0)

        if gauge:
            buffer = io.BytesIO()
            yield MAGIC
        while self.unpack.nextframe():
            while True:
                for chunk in self._apply_filter():
                    results = self._generate_bytes(chunk) if chunk.visible else (chunk,)
                    for result in results:
                        if not gauge:
                            yield result
                            continue
                        buffer.write(result)
                if self.unpack.peek[:gauge + 1] != self.unpack.trunk[:gauge + 1]:
                    break
                if not self.unpack.nextframe():
                    break
            if gauge:
                buffer.truncate(buffer.tell())
                yield Chunk(buffer.getvalue(),
                    chunk.path[:gauge], chunk.view[:gauge]).pack()
                buffer.seek(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from ....lib.patterns import formats
from .. import Deobfuscator


class deob_ps1_brackets(Deobfuscator):
    """
    PowerShell deobfuscation that removes superfluous brackets around constant
    literals, i.e. `("{0}{2}{1}")` is transformed to `"{0}{2}{1}"`. Currently,
    only integer and string constants are supported.
    """

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._sentinel = re.compile(
            RF'(?<![\w"\']{{2}})'  # this may be a function call
            RF'(\-\w+)?'  # not a function call but an argument
            RF'\(\s*({formats.integer}|{formats.ps1str})\s*\)',
            flags=re.IGNORECASE
        )

    def deobfuscate(self, data):
        while True:
            match = self._sentinel.search(data)
            if not match:
                break
            data = self._sentinel.sub(R'\1\2', data)
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import setuptools
import os
import re
import os.path
import sys

import refinery.lib.loader as loader

PREFIX = os.getenv('REFINERY_PREFIX') or ''
GITHUB = 'https://github.com/binref/refinery/'

cache_path = os.path.join(os.path.dirname(__file__), 'refinery', '__init__.pkl')
if os.path.exists(cache_path):
    os.remove(cache_path)


def normalize_name(name, separator='-'):
    return separator.join([segment for segment in name.split('_')])


def main():
    if sys.version_info < (3, 6):
        print('ERROR: Python version at least 3.6 is required.', file=sys.stderr)
        sys.exit(0xFADE)
    if sys.version_info < (3, 7):
        print('WARNING: Python 3.7 or higher is recommended, some features will be impaired.',
            file=sys.stderr)

    def magic(x):
        return '{}-win64'.format(x) if os.name == 'nt' and x == 'python-magic' else x

    requirements = [magic(l.strip()) for l in open('requirements.txt', 'r')]
    requirements = [r for r in requirements if r]

    with open('README.md', 'r', encoding='UTF8') as README:
        readme = README.read()
        readme = re.sub('(?<=\\])\\((?!\\w+://)(.*?)\\)', F'({GITHUB}blob/master/\\1)', readme)

    setuptools.setup(
        name='binary-refinery',
        version='0.1.4',
        author='Jesko Hüttenhain',
        description='A toolkit to transform and refine (mostly) binary data.',
        long_description=readme,
        long_description_content_type='text/markdown',
        url=GITHUB,
        python_requires='>=3.6',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'License :: Other/Proprietary License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: System :: Shells',
            'Topic :: Utilities'
        ],
        packages=setuptools.find_packages(
            exclude=('test*',)
        ),
        install_requires=requirements,
        entry_points={
            'console_scripts': [
                '{}{}={}:{}.run'.format(
                    PREFIX,
                    normalize_name(item.__qualname__),
                    item.__module__,
                    item.__qualname__
                )
                for item in loader.get_all_entry_points()
            ] + [
                'binre=refinery.explore:explorer'
            ]
        }
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implements various cryptographic hashing algorithms.
"""
import hashlib

from . import HashUnit


class md5(HashUnit):
    """
    Returns the MD5 hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.md5(data).digest()


class sha1(HashUnit):
    """
    Returns the SHA1 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.sha1(data).digest()


class sha224(HashUnit):
    """
    Returns the SHA224 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.sha224(data).digest()


class sha256(HashUnit):
    """
    Returns the SHA256 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.sha256(data).digest()


class sha384(HashUnit):
    """
    Returns the SHA384 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.sha384(data).digest()


class sha512(HashUnit):
    """
    Returns the SHA512 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.sha512(data).digest()


class blk224(HashUnit):
    """
    Returns the BLK224 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.blake2b(data, digest_size=28)


class blk256(HashUnit):
    """
    Returns the BLK256 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.blake2b(data, digest_size=32)


class blk384(HashUnit):
    """
    Returns the BLK384 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.blake2b(data, digest_size=48)


class blk512(HashUnit):
    """
    Returns the BLK512 Hash of the input data.
    """
    @classmethod
    def _algorithm(cls, data): return hashlib.blake2b(data, digest_size=64)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit


class hex(Unit):
    """
    Hex-decodes and encodes binary data. Non hex characters are removed from
    the input. For decoding, any odd trailing hex digits are stripped as two
    hex digits are required to represent a byte.
    """

    def reverse(self, data):
        import base64
        return base64.b16encode(data)

    def process(self, data):
        import re
        import base64
        data = re.sub(B'[^A-Fa-f0-9]+', B'', data)
        if len(data) % 2:
            data = data[:-1]
        return base64.b16decode(data, casefold=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions to help dynamically load refinery units.
"""
import pkgutil
import pkg_resources


def get_package_name():
    """
    Retrieves the toplevel package name.
    """
    root = __name__
    first_dot = root.find('.')
    if first_dot > 0:
        root = root[:first_dot]
    return root


def get_all_entry_points():
    """
    The function returns an iterator over all entry points, i.e.
    all subclasses of the `refinery.units.Entry` class.
    """
    path = [get_package_name(), 'units']
    root = __import__('.'.join(path))
    for name in path[1:]:
        root = getattr(root, name)

    def iterate(parent, *path):
        for _, name, ispkg in pkgutil.iter_modules(parent.__path__):
            try:
                module_name = '.'.join([*path, name])
                __import__(module_name)
                module = getattr(parent, name)
            except Exception:
                continue
            if ispkg:
                yield from iterate(module, *path, name)
            for attr in dir(module):
                item = getattr(module, attr)
                if getattr(item, '__module__', None) != module_name:
                    continue
                if isinstance(item, type) and issubclass(item, root.Entry) and item is not root.Entry:
                    yield item

    yield from iterate(root, *path)


def get_entry_point(name: str):
    """
    Retrieve a refinery entry point by name.
    """
    try:
        entry = pkg_resources.load_entry_point(
            get_package_name(), 'console_scripts', name).__self__
    except Exception:
        # refinery unit entry points with underscored names will be exposed with
        # these underscores replaced by dashes. If such a name is passed to the
        # loader, the following substitution ensures that it still works.
        name = name.replace('-', '_')
        for entry in get_all_entry_points():
            if getattr(entry, '__name__', None) == name:
                break
        else:
            raise AttributeError('no entry point with name "%s" was found.' % name)
    return entry


def load(name: str, *args, **kwargs):
    """
    Loads the unit specified by `name`, initialized with the given arguments
    and keyword arguments.
    """
    entry = get_entry_point(name)
    return entry(*args, **kwargs)


def load_commandline(command: str):
    """
    Returns a unit as it would be loaded from a given command line string.
    """
    import shlex
    module, *arguments = shlex.split(command)
    return load(module, *arguments)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import sliceobj

from itertools import islice, repeat, chain
from collections import deque


class pick(Unit):
    """
    Picks sequences from the array of multiple inputs. For example, `pick 0 2:`
    will return all but the second ingested input (which has index `1`).
    """
    def interface(self, argp):
        argp.add_argument(
            'slice',
            type=sliceobj,
            nargs='*',
            default=[slice(None, None)],
            help='Specify start:stop:step in Python slice syntax.'
        )
        return super().interface(argp)

    def filter(self, inputs):
        slices = deque(self.args.slice)
        discards = 0
        consumed = False

        def discardable(s):
            return s.stop and s.stop >= 0 and (s.step or 1) > 0 \
                and all(t.start >= s.stop for t in slices)

        while slices:
            s = slices.popleft()

            if not consumed:
                if not discardable(s):
                    self.log_debug(F'consumed input into buffer after {discards} skips')
                    inputs = [None] * discards + list(inputs)
                    consumed = True

            if consumed:
                yield from inputs[s]
            else:
                yield from islice(chain(repeat(None, discards), inputs), s.start, s.stop, s.step)
                discards = s.stop

        return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from .. import Deobfuscator


class deob_ps1_invoke(Deobfuscator):
    def deobfuscate(self, data):
        data = re.sub(
            R'''(\.|::)'''                    # preceeded by dot or namespace delimiter
            R'''(['"])(\w{1,200})\2'''        # quoted string (actually a method name)
            R'''(?=[\s\(\.\,\;\+\-])''',      # only if followed by certain characters
            R'\1\3', data                     # remove quotes around symbol
        )
        data = re.sub(
            '\\s{0,5}'.join([
                '[.&]', '(\\(',               # sourcing operator
                '(?:gcm|get-command)', ')?',  # potentially a get-command
                '([\'"])([-a-z]{1,100})\\2'   # string enclosing a command
                '(?(1)\\s{0,5}\\)|)',         # closing bracket for get-command
            ]), '\\3', data, flags=re.IGNORECASE
        )
        data = re.sub(
            R'''(\w{1,200})\.Invoke\s*\(''',
            R'\1(',
            data,
            flags=re.IGNORECASE
        )
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import multibin, number


class repl(Unit):
    """
    Performs a simple binary string replacement on the input data.
    """

    def interface(self, argp):
        argp.add_argument('-n', '--count', type=number, default=-1,
            help='Only replace the given number of occurrences')
        argp.add_argument('search', type=multibin,
            help='This is the search term.')
        argp.add_argument('replace', type=multibin, nargs='?', default=B'',
            help='The substitution string. Leave this empty to remove all occurrences of the search term.')
        return super().interface(argp)

    def process(self, data: bytes):
        return data.replace(
            self.args.search,
            self.args.replace,
            self.args.count
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pefile
import json
import struct

from asn1crypto import cms
from asn1crypto import x509

from ... import Unit


class pemeta(Unit):
    """
    Extract metadata from PE files, including code signing information.
    """

    def _ensure_string(self, x):
        if not isinstance(x, str):
            x = repr(x) if not isinstance(x, bytes) else x.decode(self.codec, 'backslashreplace')
        return x

    def _parse_pedict(self, bin):
        return dict((
            self._ensure_string(key),
            self._ensure_string(val)
        ) for key, val in bin.items())

    def _extract_pkcs7(self, pe, data):
        ENTRY_SECURITY = pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_SECURITY']

        pe.parse_data_directories(directories=[ENTRY_SECURITY])
        security = pe.OPTIONAL_HEADER.DATA_DIRECTORY[ENTRY_SECURITY]
        sgnoff = security.VirtualAddress + 8
        sgnend = sgnoff + security.Size

        length, revision, certtype = struct.unpack('<IHH', data[sgnoff - 8:sgnoff])
        signature = data[sgnoff:sgnend]

        self.log_debug('revision:', revision)
        self.log_debug('certtype:', certtype)
        self.log_debug('length:', length)

        if len(signature) + 8 != length:
            raise ValueError(F'Found {len(signature) + 8} bytes of signature, but length should be {length}.')

        return bytes(signature)

    def _parse_signature(self, pe, data):
        cert = self._extract_pkcs7(pe, data)
        info = {}

        try:
            signature = cms.ContentInfo.load(cert)
        except Exception as E:
            raise ValueError(F'PKCS7 parser failed with error: {E!s}')

        def tscrawl(entry, native):
            try:
                if entry['type'] == 'signing_time':
                    return entry['values']
            except KeyError:
                pass
            except TypeError:
                pass

            try:
                keys = list(entry)
                assert all(isinstance(k, str) for k in keys)
            except Exception:
                try:
                    length = len(entry)
                except TypeError:
                    return None

                for k in range(length):
                    try:
                        item = entry[k]
                    except IndexError:
                        continue
                    results = tscrawl(item, native)
                    if results:
                        return results
            else:
                for key in keys:
                    value = entry[key]
                    results = tscrawl(value, native)
                    if isinstance(results, list) and len(results) == 1:
                        result = dict(Timestamp=str(results[0]))
                        try:
                            acc = entry if native else entry.native
                            result['Timestamp Issuer'] = acc['sid']['issuer']['common_name']
                        except Exception:
                            pass
                        return result
                    elif results is not None:
                        return results

            try:
                return tscrawl(entry.native, True)
            except Exception:
                pass

        timestamps = []
        for signer in signature['content']['signer_infos']:
            timestamps.append(tscrawl(signer, False))
        if len(timestamps) == 1:
            info.update(timestamps[0])

        for entry in signature['content']['certificates']:
            cert = x509.Certificate.load(entry.dump())
            if cert.ca: continue
            tbs = cert.native['tbs_certificate']

            for extension in tbs['extensions']:
                if extension['extn_id'] == 'extended_key_usage' and 'code_signing' in extension['extn_value']:
                    tbs = dict(
                        Issuer=tbs['issuer']['common_name'],
                        Subject=tbs['subject']['common_name'],
                        Serial=F"{tbs['serial_number']:x}",
                    )
                    info.update(tbs)
                    return info

    def _parse_file_info(self, pe, data):
        pe.parse_data_directories(
            directories=[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_RESOURCE']])

        try:
            FileInfoList = pe.FileInfo
        except AttributeError:
            return None

        for FileInfo in FileInfoList:
            for FileInfoEntry in FileInfo:
                if not hasattr(FileInfoEntry, 'StringTable'):
                    continue
                for StringTableEntry in FileInfoEntry.StringTable:
                    StringTableEntryParsed = self._parse_pedict(StringTableEntry.entries)
                    StringTableEntryParsed['LangID'] = self._ensure_string(StringTableEntry.LangID)
                    return StringTableEntryParsed

    def process(self, data):
        pe = pefile.PE(data=data, fast_load=True)
        result = {}

        try:
            signature = self._parse_signature(pe, data)
        except Exception as E:
            self.log_info(F'could not extract certificate: {E!s}')
        else:
            if signature:
                result['Signature'] = signature

        try:
            file_info = self._parse_file_info(pe, data)
        except Exception as E:
            self.log_info(F'failed to obtain file info resource: {E!s}')
        else:
            if file_info:
                result['FileInfo'] = file_info

        return json.dumps(result, indent=4, ensure_ascii=False).encode(self.codec)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import JSONEncoderUnit
from .....lib.dotnet.header import DotNetHeader


class dnhdr(JSONEncoderUnit):
    """
    Expects data that has been formatted with the `BinaryFormatter` class. The
    output is a representation of the deserialized data in JSON format.
    """

    def interface(self, argp):
        argp.add_argument(
            '-r', '--resources',
            action='store_true',
            help='also parse resources'
        )
        return super().interface(argp)

    def process(self, data):
        dn = DotNetHeader(data, parse_resources=self.args.resources)
        dn = {
            'Head': dn.head,
            'Meta': dn.meta
        }

        if self.args.resources:
            dn['RSRC'] = dn.resources

        return self.to_json(dn)



import refinery as r


class carve_b64z(r.Unit):
    """
    Carves the longest base64 encoded expression and decodes it, then applies
    the `refinery.decompress` unit to the result.
    """
    def process(self, data):
        return (r.carve('-lt1', 'b64') | r.b64 | r.decompress)(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from .. import Deobfuscator, outside
from ....lib.patterns import formats


class deob_ps1_literals(Deobfuscator):
    """
    PowerShell deobfuscation that removes superfluous curly braces around
    literals that do not require it, i.e. `${variable}` is transformed to
    just `$variable`.
    """

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._sentinel = re.compile(R'\$\{(\w+)\}')

    @outside(formats.ps1str)
    def deobfuscate(self, data):
        return self._sentinel.sub(R'$\1', data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import base64

from .. import Unit


class b85(Unit):
    """
    Base85 encoding and decoding.
    """
    def reverse(self, data):
        return base64.b85encode(data)

    def process(self, data):
        return base64.b85decode(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import number


class base(Unit):
    """
    Encodes and decodes integers in arbitrary base, using the letters of the
    alphabet as an additional possible 26 digits. The largest base that can
    be represented in this manner is 36.
    """

    _DIGITS = B'0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'

    def interface(self, argp):
        argp.add_argument(
            '-e',
            '--little-endian',
            dest='byteorder',
            action='store_const',
            const='little',
            help='use little endian instead of the default big endian byte order.'
        )
        argp.add_argument(
            'base',
            type=number[2:36],
            default=0,
            nargs='?',
            help='base to be used for conversion; default value of 0 '
                 'uses common python syntax such as the 0x prefix for hexadecimal.'
        )
        argp.set_defaults(byteorder='big')
        return super().interface(argp)

    def reverse(self, data):
        self.log_info('using byte order', self.args.byteorder)
        number = int.from_bytes(data, byteorder=self.args.byteorder)

        if number == 0:
            return B'0'
        if self.args.base == 0:
            return B'0x%X' % number
        if self.args.base > len(self._DIGITS):
            raise ValueError(
                F'Only {len(self._DIGITS)} available; not enough to '
                F'encode base {self.args.base}'
            )

        def reverse_result(number):
            while number:
                yield self._DIGITS[number % self.args.base]
                number //= self.args.base

        return bytes(reversed(tuple(reverse_result(number))))

    def process(self, data):
        data = data.strip()
        number = int(data, self.args.base)
        size, rest = divmod(number.bit_length(), 8)
        if rest: size += 1
        return number.to_bytes(size, byteorder=self.args.byteorder)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from ... import Unit, RefineryPartialResult
from ....lib.argformats import multibin

from Crypto.Cipher import PKCS1_OAEP, PKCS1_v1_5
from Crypto.PublicKey import RSA


class rsa(Unit):
    """
    Implements single block RSA encryption and decryption. This unit can be used to encrypt
    and decrypt blocks generated by openssl's `rsautl` tool when using the mode `-verify`.
    When it is executed with a public key for decryption or with a private key for encryption,
    it will perform a raw RSA operation. The result of these operations are (un)padded using
    EMSA-PKCS1-v1_5.
    """
    def interface(self, argp):
        argp.add_argument('key', type=multibin, help='RSA key in PEM or DER format.')
        return super().interface(argp)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key = RSA.import_key(self.args.key)

    @property
    def blocksize(self) -> int:
        return self.key.size_in_bytes()

    def process(self, data):
        if not self.key.has_private():
            if len(data) > self.blocksize:
                raise ValueError(F'This key can only handle messages of size {self.blocksize}.')
            self.log_debug('attempting to inverse decrypt as raw blob')
            msg = int.from_bytes(data, byteorder='big')
            dec = pow(msg, self.key.e, self.key.n).to_bytes(self.blocksize, byteorder='big')
            if dec.startswith(B'\x00\x01'):
                pos = dec.find(B'\0', 2)
                if pos > 0:
                    pad = dec[2:pos]
                    if all(b == 0xFF for b in pad):
                        return dec[pos + 1:]
            raise RefineryPartialResult(
                'The decrypted message did not have a correct EMSA-PKCS1-v1_5 padding.', dec)
        result = PKCS1_v1_5.new(self.key).decrypt(data, None)
        if result is not None:
            return result
        return PKCS1_OAEP.new(self.key).decrypt(data)

    def reverse(self, data):
        if self.key.has_private():
            if len(data) > self.blocksize - 11:
                raise ValueError(F'This key can only encrypt messages of size at most {self.blocksize - 11}.')
            self.log_debug('attempting to inverse encrypt as raw blob')
            pad = self.blocksize - len(data) - 3
            msg = B'\x00\x01' + B'\xFF' * pad + B'\x00' + data
            msg = int.from_bytes(msg, byteorder='big')
            enc = pow(msg, self.key.d, self.key.n).to_bytes(self.blocksize, byteorder='big')
            return enc
        return PKCS1_OAEP.new(self.key).encrypt(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import itertools

from ...lib.tools import get_terminal_size, lookahead


class HexViewerMixin:

    def hexviewer_interface(self, argp):
        from ...lib.argformats import number
        argp.add_argument('-A', '--no-addr', dest='hexaddr', action='store_false',
            help='Do not show byte offsets in hexdump.')
        argp.add_argument('-W', '--width', metavar='N', type=number, default=0,
            help='Specify the number of hexadecimal characters to use in preview.')
        argp.add_argument('-E', '--expand', action='store_true',
            help='Do not compress sequences of identical lines in hexdump')
        return argp

    def hexaddr_size(self, total):
        addr_width = 16
        for k in range(1, 16):
            if total < (1 << (k << 2)):
                addr_width = k
                break
        return addr_width

    def hexdump(self, data, total=None):
        import re

        total = total or len(data)

        if self.args.width:
            columns = self.args.width
        else:
            # this will default to 16 byte wide output if
            # stdout is not a terminal or if its width can
            # not be determined for other reasons.
            try:
                columns = get_terminal_size() or 75
            except OSError:
                columns = 16
            else:
                if self.args.hexaddr:
                    columns -= self.hexaddr_size(total)
                    columns -= 1  # for the separator
                columns = (columns - 2) // 4

        columns = min(columns, len(data))
        lines = itertools.zip_longest(*([iter(data)] * columns))
        address_width = max(self.hexaddr_size(total), 4)
        previous = None
        prevcount = 0

        for k, (last, line) in enumerate(lookahead(lines)):
            chunk = bytes(b for b in line if b is not None)

            if not self.args.expand:
                if chunk == previous and not last:
                    prevcount += 1
                    continue
                elif prevcount > 0:
                    msg = F' repeats {prevcount} times '
                    yield F'{".":.>{address_width}}: {msg:=^{3*columns-1}}  {"":=<{columns}}'
                    prevcount = 0

            dump = ' '.join(F'{b:02X}' for b in chunk)
            read = re.sub(B'[^!-~]', B'.', chunk).decode('ascii')

            yield F'{k*columns:0{address_width}X}: {dump:<{3*columns}} {read:<{columns}}'

            if not self.args.expand:
                previous = chunk



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Cipher import ChaCha20

from . import LatinStreamCipher


class chacha(LatinStreamCipher, cipher=ChaCha20):
    """
    ChaCha20 and XChaCha20 encryption and decryption. For ChaCha20, the IV
    (nonce) must be 8 or 12 bytes long; for XChaCha20, choose an IV which is
    24 bytes long. The default IV value is the string value `REFINERY`.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Routines to help interpret large binary buffers as arrays of numbers, stored
as consecutive sequences of bytes, all with the same length and byte order.
"""


def unpack(data: bytes, blocksize, little_endian=True):
    """
    Returns an iterable of integers which have been unpacked from the given `data`
    buffer as chunks of `blocksize` many bytes.
    """
    if blocksize == 1:
        return data
    if blocksize in (2, 4, 8):
        try:
            import numpy
        except ModuleNotFoundError:
            numpy = None
        order = '<' if little_endian else '>'
        count = len(data) // blocksize
        if numpy:
            dtype = numpy.dtype(F'{order}u{blocksize}')
            return (int(x) for x in numpy.frombuffer(data, dtype, count))
        else:
            import struct
            scode = {2: 'H', 4: 'L', 8: 'Q'}[blocksize]
            return struct.unpack(F'{order}{count}{scode}', data[:count * blocksize])
    else:
        blocks = zip(*([iter(data)] * blocksize))
        byteorder = ('big', 'little')[little_endian]
        return (int.from_bytes(block, byteorder) for block in blocks)


def pack(data, blocksize, little_endian=True):
    """
    Returns a bytes object which contains the packed representation of the
    integers in `data`, where each item is encoded using `blocksize` many
    bytes. The numbers are assumed to fit this encoding.
    """
    if blocksize == 1:
        return bytes(data)
    if blocksize in (2, 4, 8):
        try:
            import numpy
        except ModuleNotFoundError:
            numpy = None
        order = '<' if little_endian else '>'
        if numpy:
            dtype = numpy.dtype(F'{order}u{blocksize}')
            return numpy.fromiter(data, dtype).tobytes()
        else:
            import struct
            scode = {2: 'H', 4: 'L', 8: 'Q'}[blocksize]
            return struct.pack(F'{order}{len(data)}{scode}', *data)
    byteorder = ('big', 'little')[little_endian]
    return B''.join(number.to_bytes(blocksize, byteorder) for number in data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A cross platform interface to libmagic.
"""
try:
    from winmagic import magic
except ModuleNotFoundError:
    import os
    if os.name == 'nt':
        # Attempting to import magic on Windows without winmagic being
        # installed may result in an uncontrolled crash.
        magic = None
    else:
        try:
            import magic
        except ModuleNotFoundError:
            magic = None


def magicparse(data, *args, **kwargs):
    if magic:
        data = bytes(data) if not isinstance(data, bytes) else data
        return magic.Magic(*args, **kwargs).from_buffer(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from ... import RefineryPartialResult
from . import KeyDerivation

from Crypto.Cipher import DES

__all__ = ['DESDerive']


class DESDerive(KeyDerivation):
    """
    Implements the same functionality as `DES_string_to_key` in OpenSSL. It
    converts a string to an 8 byte DES key with odd byte parity, per FIPS
    specification. This is not a modern key derivation function.
    """

    _DEFAULT_SIZE = 8

    _TABLE = bytearray((
        0x01, 0x01, 0x02, 0x02, 0x04, 0x04, 0x07, 0x07,
        0x08, 0x08, 0x0B, 0x0B, 0x0D, 0x0D, 0x0E, 0x0E,
        0x10, 0x10, 0x13, 0x13, 0x15, 0x15, 0x16, 0x16,
        0x19, 0x19, 0x1A, 0x1A, 0x1C, 0x1C, 0x1F, 0x1F,
        0x20, 0x20, 0x23, 0x23, 0x25, 0x25, 0x26, 0x26,
        0x29, 0x29, 0x2A, 0x2A, 0x2C, 0x2C, 0x2F, 0x2F,
        0x31, 0x31, 0x32, 0x32, 0x34, 0x34, 0x37, 0x37,
        0x38, 0x38, 0x3B, 0x3B, 0x3D, 0x3D, 0x3E, 0x3E,
        0x40, 0x40, 0x43, 0x43, 0x45, 0x45, 0x46, 0x46,
        0x49, 0x49, 0x4A, 0x4A, 0x4C, 0x4C, 0x4F, 0x4F,
        0x51, 0x51, 0x52, 0x52, 0x54, 0x54, 0x57, 0x57,
        0x58, 0x58, 0x5B, 0x5B, 0x5D, 0x5D, 0x5E, 0x5E,
        0x61, 0x61, 0x62, 0x62, 0x64, 0x64, 0x67, 0x67,
        0x68, 0x68, 0x6B, 0x6B, 0x6D, 0x6D, 0x6E, 0x6E,
        0x70, 0x70, 0x73, 0x73, 0x75, 0x75, 0x76, 0x76,
        0x79, 0x79, 0x7A, 0x7A, 0x7C, 0x7C, 0x7F, 0x7F,
        0x80, 0x80, 0x83, 0x83, 0x85, 0x85, 0x86, 0x86,
        0x89, 0x89, 0x8A, 0x8A, 0x8C, 0x8C, 0x8F, 0x8F,
        0x91, 0x91, 0x92, 0x92, 0x94, 0x94, 0x97, 0x97,
        0x98, 0x98, 0x9B, 0x9B, 0x9D, 0x9D, 0x9E, 0x9E,
        0xA1, 0xA1, 0xA2, 0xA2, 0xA4, 0xA4, 0xA7, 0xA7,
        0xA8, 0xA8, 0xAB, 0xAB, 0xAD, 0xAD, 0xAE, 0xAE,
        0xB0, 0xB0, 0xB3, 0xB3, 0xB5, 0xB5, 0xB6, 0xB6,
        0xB9, 0xB9, 0xBA, 0xBA, 0xBC, 0xBC, 0xBF, 0xBF,
        0xC1, 0xC1, 0xC2, 0xC2, 0xC4, 0xC4, 0xC7, 0xC7,
        0xC8, 0xC8, 0xCB, 0xCB, 0xCD, 0xCD, 0xCE, 0xCE,
        0xD0, 0xD0, 0xD3, 0xD3, 0xD5, 0xD5, 0xD6, 0xD6,
        0xD9, 0xD9, 0xDA, 0xDA, 0xDC, 0xDC, 0xDF, 0xDF,
        0xE0, 0xE0, 0xE3, 0xE3, 0xE5, 0xE5, 0xE6, 0xE6,
        0xE9, 0xE9, 0xEA, 0xEA, 0xEC, 0xEC, 0xEF, 0xEF,
        0xF1, 0xF1, 0xF2, 0xF2, 0xF4, 0xF4, 0xF7, 0xF7,
        0xF8, 0xF8, 0xFB, 0xFB, 0xFD, 0xFD, 0xFE, 0xFE
    ))

    def process(self, password):
        key = bytearray(8)

        for i, c in enumerate(password):
            if ((i % 16) < 8):
                key[i % 8] ^= (c << 1) & 0xFF
            else:
                c = (((c << 4) & 0xf0) | ((c >> 4) & 0x0f))
                c = (((c << 2) & 0xcc) | ((c >> 2) & 0x33))
                c = (((c << 1) & 0xaa) | ((c >> 1) & 0x55))
                key[7 - (i % 8)] ^= c

        n = len(password)

        key[:] = (self._TABLE[b] for b in key)
        des = DES.new(key, DES.MODE_CBC, iv=key)
        key = des.decrypt(password.ljust(n + 8 - n % 8, b'\0'))[-8:]
        key = bytes((self._TABLE[b] for b in key))

        if self.args.size > 8:
            raise RefineryPartialResult('DESDerive can provide at most 8 bytes.', partial=key)
        return key[:self.args.size]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.patterns import defanged, indicators


class defang(Unit):
    """
    Defangs all domains and ipv4 addresses in the input data by replacing the
    last dot in the expression by `[.]`. For example, `127.0.0.1` will be
    replaced by `127.0.0[.]1`.
    """

    WHITELIST = [
        B'wscript.shell',
    ]

    def interface(self, argp):
        argp.add_argument('-q', '--quote', action='store_true', help='Wrap all indicators in backticks for markdown code.')
        argp.add_argument('-u', '--url-only', action='store_true', help='Only defang URLs, do not look for domains or IPs.')
        argp.add_argument('-d', '--dot-only', dest='protocol', action='store_false', help='Do not escape the protocol colon in URLs.')
        return super().interface(argp)

    def _quote(self, word):
        return word if not self.args.quote else B'`%s`' % word

    def reverse(self, data):
        def refang(hostname):
            return hostname.group(0).replace(B'[.]', B'.')
        data = defanged.hostname.sub(refang, data)
        data = data.replace(B'[:]//', B'://')
        return data

    def process(self, data):
        def replace_hostname(hostname, match=True):
            if match:
                return self._quote(replace_hostname(hostname.group(0), False))
            self.log_info('replace:', hostname)
            host = hostname.rsplit(B':')[0].lower()
            if host in self.WHITELIST:
                return hostname
            return B'[.]'.join(hostname.rsplit(B'.', 1))

        def replace_url(url):
            if not url:
                return url
            sep = B'[:]//' if self.args.protocol else B'://'
            self.log_info('replace:', url)
            p, q = url.split(B'://')
            q = q.split(B'/', 1)
            q[0] = replace_hostname(q[0], False)
            q = B'/'.join(q)
            return self._quote(p + sep + q)

        analyze = indicators.url.split(data)
        analyze[1::2] = [replace_url(t) for t in analyze[1::2]]

        if not self.args.url_only:
            analyze[0::2] = [
                indicators.hostname.sub(replace_hostname, t)
                for t in analyze[0::2]
            ]

        return B''.join(analyze)
